"""Server wrapper."""

from argparse import ArgumentParser, Namespace
from logging import DEBUG, INFO, WARNING, basicConfig, getLogger
from pathlib import Path
from subprocess import Popen

from dzdsu.constants import JSON_FILE
from dzdsu.server import load_servers


__all__ = ['main']


LOGGER = getLogger('dzdsw')


def get_args(description: str = __doc__) -> Namespace:
    """Return the parsed command line arguments."""

    parser = ArgumentParser(description=description)
    parser.add_argument('server', help='the name of the server to start')
    parser.add_argument(
        '-f', '--servers-file', type=Path, default=JSON_FILE, metavar='file',
        help='servers JSON file path'
    )
    parser.add_argument(
        '-F', '--fork', action='store_true',
        help='fork server process to background'
    )
    parser.add_argument(
        '-p', '--store-pid', action='store_true',
        help="store the process' PID in its PID file"
    )
    parser.add_argument(
        '-v', '--verbose', action='store_true', help='verbose logging'
    )
    parser.add_argument(
        '-d', '--debug', action='store_true', help='debug logging'
    )
    return parser.parse_args()


def main() -> int:
    """Starts the DayZ server."""

    args = get_args()
    basicConfig(
        level=DEBUG if args.debug else INFO if args.verbose else WARNING
    )
    servers = load_servers(args.servers_file)
    env = None

    try:
        server = servers[args.server]
    except KeyError:
        LOGGER.error('No such server: %s', args.server)
        return 2

    if server.update_lockfile.exists():
        LOGGER.error('Server is currently updating.')
        return 3

    server.update_hashes()
    proc = Popen(server.command, cwd=server.base_dir, env=env)

    if args.store_pid:
        server.pid = proc.pid

    if args.fork:
        return 0

    return proc.wait()



"""Watchdog to detect server and mod updates."""


__all__ = ['hash_changed']


def hash_changed(old: dict[str, str], new: dict[str, str]) -> bool:
    """Returns True iff the hashes are considered equal."""

    for key, value in old.items():
        if (new_value := new.get(key)) is not None and new_value != value:
            return True

    for key in new.keys():
        if key not in old:
            return True

    return False



"""Server utility."""

from argparse import ArgumentParser, Namespace
from logging import DEBUG, INFO, WARNING, basicConfig, getLogger
from os import kill, name
from pathlib import Path
from signal import SIGINT
from tempfile import TemporaryDirectory

from dzdsu.constants import JSON_FILE, SHUTDOWN_MESSAGE
from dzdsu.hash import hash_changed
from dzdsu.mods import print_mods
from dzdsu.server import Server, load_servers
from dzdsu.update import Updater


__all__ = ['main']


LOGGER = getLogger('dzdsu')


def get_args(description: str = __doc__) -> Namespace:
    """Return the parsed command line arguments."""

    parser = ArgumentParser(description=description)
    parser.add_argument('server', help='the server to operate on')
    parser.add_argument(
        '-f', '--servers-file', type=Path, default=JSON_FILE, metavar='file',
        help='servers JSON file path'
    )
    parser.add_argument(
        '-C', '--clean-mods', action='store_true', help='remove unused mods'
    )
    parser.add_argument(
        '-U', '--update', metavar='steam_user',
        help='update server and/or mods'
    )
    parser.add_argument(
        '-s', '--update-server', action='store_true', help='update server'
    )
    parser.add_argument(
        '-m', '--update-mods', action='store_true',
        help="update the server's mods"
    )
    parser.add_argument(
        '-F', '--fix-paths', action='store_true', help='fix mod file paths'
    )
    parser.add_argument(
        '-K', '--install-keys', action='store_true', help='install mod keys'
    )
    parser.add_argument(
        '--overwrite', action='store_true', help="overwrite existing key files"
    )
    parser.add_argument(
        '-M', '--list-mods', action='store_true', help="list the server's mods"
    )
    parser.add_argument(
        '-S', '--list-server-mods', action='store_true',
        help="list the server's server mods"
    )
    parser.add_argument(
        '-I', '--installed-mods', action='store_true',
        help='list installed mods'
    )
    parser.add_argument(
        '-T', '--shutdown', action='store_true',
        help="shutdown the server if it needs a restart"
    )
    parser.add_argument(
        '-N', '--needs-restart', action='store_true',
        help="check whether the server needs a restart"
    )
    parser.add_argument(
        '--message', default=SHUTDOWN_MESSAGE, metavar='template',
        help='RCon message template for messages to users'
    )
    parser.add_argument(
        '-t', '--gracetime', type=int, default=120, metavar='seconds',
        help='grace time to wait before server shutdown'
    )
    parser.add_argument(
        '-v', '--verbose', action='store_true', help='verbose logging'
    )
    parser.add_argument(
        '-d', '--debug', action='store_true', help='debug logging'
    )
    return parser.parse_args()


def clean_mods(server: Server) -> None:
    """Remove unused mods."""

    for installed_mod in server.unused_mods:
        LOGGER.info('Removing unused mod: %s', installed_mod.mod)
        installed_mod.remove()


def install_keys(server: Server) -> None:
    """Installs the keys for all mods of the server."""

    for installed_mod in server.installed_mods:
        for key in installed_mod.bikeys:
            if (installed := server.base_dir / 'keys' / key.name).exists():
                LOGGER.info('Key "%s" already installed.', key.name)
                continue

            with key.open('rb') as src, installed.open('wb') as dst:
                dst.write(src.read())


def fix_mod_paths(server: Server) -> None:
    """Fix paths of the server mods."""

    for installed_mod in server.installed_mods:
        LOGGER.info('Fixing paths of: %s', installed_mod.mod)
        installed_mod.fix_paths()


def _update(server: Server, args: Namespace) -> None:
    """Perform server and mod updates."""

    updater = Updater(server, args.update)

    if args.update_server:
        updater.update_server()

    if args.update_mods:
        updater.update_mods()

    with server.update_lockfile:
        updater()


def needs_update_nt(server: Server, args: Namespace) -> bool:
    """Returns True iff there is an update available."""

    with TemporaryDirectory() as temp_dir:
        _update(copy := server.chdir(Path(temp_dir)), args)
        return hash_changed(server.hashes, copy.hashes)


def pre_update_shutdown(server: Server, args: Namespace) -> bool:
    """Conditionally shutdown server before update."""

    if not needs_update_nt(server, args):
        LOGGER.info('No update required.')
        return False

    if not shutdown(server, args):
        LOGGER.warning('Could not shutdown server prior to update.')
        return False

    return True


def update(server: Server, args: Namespace) -> None:
    """Updates the server."""

    # Windows systems cannot override files that are in use by a process.
    # So we need to shut the server down *before* the update.
    if name == 'nt' and not pre_update_shutdown(server, args):
        return

    _update(server, args)


def shutdown(server: Server, args: Namespace) -> int:
    """Shut down the server iff it needs a restart."""

    if (pid := server.pid) is None:
        LOGGER.error('No PID found for server.')
        return 2

    LOGGER.info('Updates detected. Notifying users.')

    if not server.notify_shutdown(args.message, grace_time=args.grace_time):
        LOGGER.error('Could not notify users about shutdown.')
        return 3

    LOGGER.info(f'Kicking remaining users.')
    server.kick_all('Server restart.')
    LOGGER.info(f'Stopping server.')

    try:
        kill(pid, SIGINT)
    except ProcessLookupError:
        LOGGER.error('Could not find process with PID: %i', pid)
        return 4

    return 0


def main() -> int:
    """Update mods."""

    args = get_args()
    basicConfig(
        level=DEBUG if args.debug else INFO if args.verbose else WARNING
    )
    servers = load_servers(args.servers_file)

    try:
        server = servers[args.server]
    except KeyError:
        LOGGER.error('No such server: %s', args.server)
        return 2

    if args.clean_mods:
        clean_mods(server)

    if args.update:
        update(server, args)

    if args.fix_paths:
        fix_mod_paths(server)

    if args.install_keys:
        install_keys(server)

    if args.list_mods:
        print_mods(server.mods)

    if args.list_server_mods:
        print_mods(server.server_mods)

    if args.installed_mods:
        print_mods(sorted(map(
            lambda installed_mod: installed_mod.mod, server.installed_mods
        )))

    if args.shutdown:
        return shutdown(server, args)

    if args.needs_restart:
        return 0 if server.needs_restart else 1

    return 0



"""Server representation."""

from __future__ import annotations
from configparser import SectionProxy
from hashlib import sha1
from itertools import chain
from json import dump, load
from pathlib import Path
from time import sleep
from typing import Any, Iterator, NamedTuple

from rcon.battleye import Client

from dzdsu.constants import BATTLEYE_GLOB
from dzdsu.constants import DAYZ_SERVER_APP_ID
from dzdsu.constants import JSON_FILE
from dzdsu.constants import MODS_DIR
from dzdsu.constants import SERVER_EXECUTABLE
from dzdsu.constants import SHUTDOWN_MESSAGE
from dzdsu.hash import hash_changed
from dzdsu.lockfile import LockFile
from dzdsu.mods import Mod, ModMetadata, InstalledMod, mods_str
from dzdsu.params import ServerParams
from dzdsu.parsers import parse_battleye_cfg, parse_server_cfg


__all__ = ['Server', 'load_servers']


class Server(NamedTuple):
    """A server."""

    name: str
    app_id: int
    base_dir: Path
    executable: Path
    mods: list[Mod]
    server_mods: list[Mod]
    params: ServerParams

    @classmethod
    def from_json(cls, name: str, json: dict):
        """Creates a Server instance from a JSON-ish dict."""
        return cls(
            name,
            json.get('appId', DAYZ_SERVER_APP_ID),
            Path(json['basedir']),
            Path(json.get('executable', SERVER_EXECUTABLE)),
            [Mod.from_value(mod) for mod in (json.get('mods') or [])],
            [Mod.from_value(mod) for mod in (json.get('serverMods') or [])],
            ServerParams.from_json(json.get('params') or {})
        )

    @property
    def executable_path(self) -> Path:
        """Returns the full command for running the server."""
        if self.executable.is_absolute():
            return self.executable

        return self.base_dir / self.executable

    @property
    def sha1sum(self) -> str:
        """Returns the SHA-1 checksum."""
        with self.executable_path.open('rb') as file:
            return sha1(file.read()).hexdigest()

    @property
    def executable_args(self) -> Iterator[str]:
        """Yields arguments for the server executable."""
        yield from self.params.executable_args
        yield f'-BEpath={self.battleye_dir}'

        if mods := mods_str(mod for mod in self.mods if mod.enabled):
            yield f'-mod={mods}'

        if mods := mods_str(mod for mod in self.server_mods if mod.enabled):
            yield f'-serverMod={mods}'

    @property
    def command(self) -> list[str]:
        """Returns the full command for running the server."""
        return [str(self.executable_path), *self.executable_args]

    @property
    def mods_dir(self) -> Path:
        """Returns the server's mods directory."""
        return self.base_dir / MODS_DIR

    @property
    def battleye_dir(self) -> Path:
        """Returns the profile directory."""
        return self.base_dir / 'battleye'

    @property
    def battleye_cfg_file(self) -> Path:
        """Returns the BattlEye RCon config file."""
        for path in self.battleye_dir.glob(BATTLEYE_GLOB):
            if path.is_file():
                return path

        raise FileNotFoundError(BATTLEYE_GLOB)

    @property
    def battleye_cfg(self) -> dict[str, str | int | bool]:
        """Returns the BattlEye RCon configuration."""
        with self.battleye_cfg_file.open('r', encoding='utf-8') as file:
            return dict(parse_battleye_cfg(file))

    @property
    def installed_mods_metadata(self) -> Iterator[ModMetadata]:
        """Yields metadata of installed mods."""
        for filename in self.mods_dir.glob('*/meta.cpp'):
            yield ModMetadata.from_file(filename)

    @property
    def config_file(self) -> Path:
        """Returns the config file path."""
        return self.base_dir / 'serverDZ.cfg'

    @property
    def config(self) -> SectionProxy:
        """Returns the configuration settings."""
        with self.config_file.open('r', encoding='utf-8') as file:
            return parse_server_cfg(file)

    @property
    def installed_mods(self) -> Iterator[InstalledMod]:
        """Yields installed mods."""
        for meta in self.installed_mods_metadata:
            yield InstalledMod(meta.publishedid, self.base_dir)

    @property
    def used_mods(self) -> Iterator[InstalledMod]:
        """Yields used mods."""
        used_ids = {mod.id for mod in chain(self.mods, self.server_mods)}

        for meta in self.installed_mods_metadata:
            if meta.publishedid in used_ids:
                yield InstalledMod(meta.publishedid, self.base_dir)

    @property
    def unused_mods(self) -> Iterator[InstalledMod]:
        """Yields unused mods."""
        used_ids = {mod.id for mod in chain(self.mods, self.server_mods)}

        for meta in self.installed_mods_metadata:
            if meta.publishedid not in used_ids:
                yield InstalledMod(meta.publishedid, self.base_dir)

    @property
    def hashes_file(self) -> Path:
        """Returns a hashes file."""
        return self.base_dir / '.hashes.json'

    @property
    def hashes(self) -> dict[str, str]:
        """Returns the server's and its mods' hashes."""
        return {
            'server': self.sha1sum,
            **{str(mod.id): mod.sha1sum for mod in self.used_mods}
        }

    @property
    def needs_restart(self) -> bool:
        """Checks whether the server needs a restart."""
        return hash_changed(self.hashes, self.load_hashes())

    @property
    def pid_file(self) -> Path:
        """Returns the path to the PID file."""
        return self.base_dir / '.pidfile'

    @property
    def pid(self) -> int | None:
        """Reads the server's PID."""
        try:
            with self.pid_file.open('rb') as file:
                return load(file)
        except FileNotFoundError:
            return None

    @pid.setter
    def pid(self, pid: int) -> None:
        """Sets the server's PID."""
        with self.pid_file.open('w', encoding='utf-8') as file:
            dump(pid, file)

    @property
    def update_lockfile(self) -> LockFile:
        """Returns the path to the update lock file."""
        return LockFile(self.base_dir / '.update.lck')

    def chdir(self, base_dir: Path) -> Server:
        """Returns a server copy with a changed base dir."""
        return Server(
            self.name, self.app_id, base_dir, self.executable,
            self.mods, self.server_mods, self.params
        )

    def update_hashes(self) -> None:
        """Updates the hashes file."""
        with self.hashes_file.open('w', encoding='utf-8') as file:
            dump(self.hashes, file)

    def load_hashes(self) -> dict[str, str]:
        """Loads hashes for the server."""
        try:
            with self.hashes_file.open('rb') as file:
                return load(file)
        except FileNotFoundError:
            return {}

    def rcon(self, timeout: float | None = None):
        """Returns an RCon client."""
        return Client(
            (config := self.battleye_cfg).get('RConIP', '127.0.0.1'),
            config.get('RConPort', 2302),
            passwd=config['RConPassword'],
            timeout=timeout
        )

    def _notify_shutdown(self, template: str, grace_time: int) -> None:
        """Notify users about shutdown."""
        with self.rcon(timeout=1) as rcon:
            for passed in range(grace_time):
                rcon.run(f'say -1 {template.format(grace_time - passed)}')
                sleep(1)
                rcon.login(rcon.passwd)

    def notify_shutdown(
            self,
            template: str = SHUTDOWN_MESSAGE,
            grace_time: int = 120
    ) -> bool:
        """Notify users about shutdown."""
        if grace_time <= 0:
            return True

        try:
            self._notify_shutdown(template, grace_time)
        except (ConnectionRefusedError, TimeoutError):
            return False

    def kick(self, player: int, reason: str) -> None:
        """Kicks the respective player."""
        with self.rcon(timeout=1) as rcon:
            rcon.run(f'kick {player} {reason}')

    def kick_all(self, reason: str) -> None:
        """Kick all players."""
        with self.rcon(timeout=1) as rcon:
            for player in range(self.config.getint('maxPlayers')):
                rcon.run(f'kick {player} {reason}')


def load_servers_json(file: Path) -> dict[str, Any]:
    """Loads servers from a JSON file."""

    with file.open('rb') as json:
        return load(json)


def load_servers(file: Path = JSON_FILE) -> dict[str, Server]:
    """Loads servers."""

    return {
        name: Server.from_json(name, json)
        for name, json in load_servers_json(file).items()
    }



"""Modifications from the Steam workshop."""

from __future__ import annotations
from hashlib import sha1
from logging import getLogger
from pathlib import Path
from shutil import rmtree
from typing import Iterable, Iterator, NamedTuple, Optional

from dzdsu.constants import ITALIC
from dzdsu.constants import LINK
from dzdsu.constants import MODS_DIR
from dzdsu.constants import WORKSHOP_URL


__all__ = ['Mod', 'ModMetadata', 'InstalledMod', 'mods_str', 'print_mods']


class Mod(NamedTuple):
    """A server mod."""

    id: int
    name: Optional[str] = None
    enabled: bool = True

    def __str__(self) -> str:
        return LINK.format(url=self.url, text=self.name or self.id)

    @classmethod
    def from_id(cls, ident: int, *, name: Optional[str] = None) -> Mod:
        """Creates a mod from an ID."""
        if ident == 0:
            raise ValueError(f'Invalid mod ID: {ident}')

        if ident < 0:
            return cls(abs(ident), name, enabled=False)

        return cls(ident, name)

    @classmethod
    def from_json(cls, json: dict[str, int | str]) -> Mod:
        """Creates a mod from a JSON-ish dict."""
        return cls.from_id(json['id'], name=json.get('name'))

    @classmethod
    def from_value(cls, value: int | dict[str, int | str]) -> Mod:
        """Creates a mod from an int or JSON value."""
        if isinstance(value, int):
            return cls.from_id(value)

        if isinstance(value, dict):
            return cls.from_json(value)

        raise TypeError(f'Cannot create mod from: {value} ({type(value)})')

    @property
    def path(self) -> Path:
        """Returns the relative path to the local mod directory."""
        return MODS_DIR / str(self.id)

    @property
    def url(self) -> str:
        """Returns the Steam Workshop URL."""
        return WORKSHOP_URL.format(self.id)


class ModMetadata(NamedTuple):
    """Metadata of a mod."""

    protocol: int
    publishedid: int
    name: str
    timestamp: int

    @classmethod
    def from_dict(cls, dct: dict) -> ModMetadata:
        """Creates mod metadata from the given dict."""
        return cls(
            int(dct['protocol']),
            int(dct['publishedid']),
            dct['name'].strip('"'),
            int(dct['timestamp'])
        )

    @classmethod
    def from_lines(cls, lines: Iterable[str]) -> ModMetadata:
        """Creates mod metadata from the given lines."""
        return cls.from_dict({
            key: value.rstrip(';') for key, value in (
                line.split(' = ') for line in (
                    line.strip() for line in lines
                ) if line
            )
        })

    @classmethod
    def from_file(cls, filename: Path) -> ModMetadata:
        """Reads the mod metadata from the given file."""
        with filename.open('r', encoding='utf-8') as file:
            return cls.from_lines(file)


class InstalledMod(NamedTuple):
    """Represents an installed mod."""

    id: int
    base_dir: Path

    @property
    def mod(self) -> Mod:
        """Returns a Mod object."""
        return Mod(self.id, self.metadata.name)

    @property
    def path(self) -> Path:
        """Returns the relative path to the local mod directory."""
        return self.base_dir / MODS_DIR / str(self.id)

    @property
    def addons(self) -> Path:
        """Returns the path to the addons directory."""
        return self.path / 'addons'

    @property
    def keys(self) -> Path:
        """Returns the path to the keys directory."""
        return self.path / 'keys'

    @property
    def metadata_file(self) -> Path:
        """Returns the path to the metadata file."""
        return self.path / 'meta.cpp'

    @property
    def metadata(self) -> ModMetadata:
        """Returns the mod metadata."""
        return ModMetadata.from_file(self.metadata_file)

    @property
    def sha1sum(self) -> str:
        """Returns the SHA-1 checksum."""
        with self.metadata_file.open('rb') as file:
            return sha1(file.read()).hexdigest()

    @property
    def pbos(self) -> Iterator[Path]:
        """Yields paths to the .pbo files."""
        return self.addons.glob('*.pbo')

    @property
    def bikeys(self) -> Iterator[Path]:
        """Yields paths to the *.bikey files."""
        return self.keys.glob('*.bikey')

    def fix_paths(self) -> None:
        """Links paths to lower-case."""
        if (addons := self.path / 'Addons').is_dir():
            link_to_lowercase(addons)

        if (keys := self.path / 'Keys').is_dir():
            link_to_lowercase(keys)

        for pbo in self.pbos:
            link_to_lowercase(pbo)

    def remove(self) -> None:
        """Removes this mod."""
        rmtree(self.path)


def link_to_lowercase(path: Path) -> None:
    """Creates a symlink with the path names in lower case."""

    if (filename := path.name) == (lower := filename.lower()):
        return

    if (symlink := path.parent / lower).exists():
        return

    getLogger(__file__).debug('Linking "%s" to "%s".', filename, symlink)
    symlink.symlink_to(filename)


def mods_str(mods: Iterable[Mod], sep: str = ';') -> str:
    """Returns a string representation of the given mods."""

    return sep.join(str(mod.path) for mod in mods)


def print_mods(mods: Iterable[Mod]) -> None:
    """Lists the respective mods."""

    for mod in mods:
        print(mod if mod.enabled else ITALIC.format(mod))



"""Lock file implementation."""

from pathlib import Path


__all__ = ['LockFile']


class LockFile:
    """A lock file."""

    def __init__(self, file: Path):
        self.file = file

    def __getattr__(self, item):
        return getattr(self.file, item)

    def __enter__(self):
        if self.file.exists():
            raise FileExistsError()

        with self.file.open('wb') as file:
            file.write(b'locked')

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.file.unlink()
